#pragma once

#include <algorithm>

#include <amdis/common/IndexSeq.hpp>
#include <amdis/common/Math.hpp>
#include <amdis/common/Mpl.hpp>
#include <amdis/common/Concepts.hpp>
#include <amdis/common/ScalarTypes.hpp>

namespace AMDiS
{
  namespace Concepts
  {
    namespace Definition
    {
      struct HasFunctorOrder
      {
        template <class F, std::size_t... I>
        auto requires_(F&& f, Indices<I...>) -> decltype( order(f, int(I)...) );
      };
    }

    template <class F, std::size_t N>
    constexpr bool HasFunctorOrder = models<Definition::HasFunctorOrder(F, MakeSeq_t<N>)>;
  }

  namespace Operation
  {
    /** \addtogroup operations
     *  @{
     **/

    /// Functor representing a static constant value
    template <class T, T value>
    struct StaticConstant
    {
      template <class... Ts>
      constexpr T operator()(Ts const&... /*args*/) const
      {
        return value;
      }
    };

    using Zero = StaticConstant<int, 0>;
    using One = StaticConstant<int, 1>;

    template <class T, T value, class... Int>
    constexpr int order(StaticConstant<T,value> const&, Int... /*orders*/)
    {
      return 0;
    }

    template <class T, T value, std::size_t J>
    constexpr auto partial(StaticConstant<T,value>, index_t<J>)
    {
      return Zero{};
    }

    // -------------------------------------------------------------------------

    /// (Unary-)Functor representing the identity
    struct Id
    {
      template <class T>
      constexpr T const& operator()(T const& x) const
      {
        return x;
      }

      friend constexpr int order(Id, int d)
      {
        return d;
      }

      friend auto partial(Id, index_t<0>)
      {
        return One{};
      }
    };

    // -------------------------------------------------------------------------

    /// Functor representing a constant value
    template <class T>
    struct Constant
    {
      constexpr Constant(T value)
        : value_(value)
      {}

      template <class... Ts>
      constexpr T const& operator()(Ts const&... /*args*/) const
      {
        return value_;
      }

    private:
      T value_;
    };

    template <class T, class... Int>
    constexpr int order(Constant<T> const&, Int... /*orders*/)
    {
      return 0;
    }

    template <class T, std::size_t J>
    constexpr auto partial(Constant<T> const&, index_t<J>)
    {
      return Zero{};
    }

    // -------------------------------------------------------------------------

    template <std::size_t I>
    struct Arg
    {
      template <class... Ts>
      constexpr auto&& operator()(Ts&&... args) const
      {
        return std::get<I>(std::forward_as_tuple(std::forward<Ts>(args)...));
      }
    };

    template <std::size_t I, class... Int>
    constexpr int order(Arg<I> const&, Int... orders)
    {
      return std::get<I>(std::forward_as_tuple(orders...));
    }

    template <std::size_t I, std::size_t J>
    constexpr auto partial(Arg<I>, index_t<J>)
    {
      return StaticConstant<int,(I==J ? 1 : 0)>{};
    }

    // -------------------------------------------------------------------------

    template <std::size_t I>
    struct Get
    {
      template <class T, int N>
      constexpr T const& operator()(Dune::FieldVector<T,N> const& vec) const
      {
        return vec[I];
      }

      friend constexpr int order(Get, int d)
      {
        return d;
      }
    };

    struct Get_
    {
      explicit constexpr Get_(std::size_t i)
        : i_(i)
      {}

      template <class T, int N>
      constexpr T const& operator()(Dune::FieldVector<T,N> const& vec) const
      {
        return vec[i_];
      }

      friend constexpr int order(Get_, int d)
      {
        return d;
      }

      std::size_t i_;
    };

    /** @} **/

  } // end namespace Operation
} // end namespace AMDiS
